/* interp.c  s-code virtual machine

	som-v16 add file i/o 			7 Nov 2005
	som-v16u flat sx-code			1 Mar 2007
	som-v31  faster sx-code			19 Aug 2007

 */

#include "compile.h"

#define		PRIVATE 	static

// stack macro
#define   TOS		tos
#define   NEXT      SS[sp]
#define   ARG       F[a]
#define	  pop1      sp--

// inline is 4x faster than push/pop functions
#define   pop()		SS[sp]; sp--
#define   push(x)   sp++; SS[sp] = tos; tos = (x)

/* run time environment
   define code, data, stack segments here

   system area  1 .. MAXSYS-1
   1..80   tokstring-som
*/

int M[MAXMEM];				// data and code segment
int SS[MAXSS+100];			// stack segment
int DS = MAXSYS;			// current free data segment
FILE *filep[MAXFILE];		// store file pointer

int load_ads;				// base ads for fun "loadfile"
int xop_ads, xarg_ads;		// base ads for XOP[] XARG[]

int *xop = &M[MAXDS];		// decoded instruction
int *xarg = &M[MAXDS+N_CS]; // decoded argument

int fp = 1, sp = 2;			// current state of computation
int tos;					// top of stack register
int *F;						// pointer to stack frame

//int IP;					// current ip
//int flag = 0;
//int cnt = 0;

// switch xop[], xarg[] between user and system
#define  user_cs()  xop=&M[M[xop_ads]]; xarg=&M[M[xarg_ads]]
#define  sys_cs()   xop=&M[MAXDS]; xarg=&M[MAXDS+N_CS]

// -------------------------

#ifdef PROF

int icnt[100], lcnt[10000];

void stat(void){
	icnt[xop[IP]]++;
	if(IP < 10000) lcnt[IP]++;
}

void report(void){
	int i, s;
	s = 0;
	for(i=1; i<EOC; i++)
		if(icnt[i] != 0){
//			prCode2(stdout,i);
			printf("%d %d\n",i,icnt[i]);
			s += icnt[i];
		}
	printf("total %d\n",s);

	s = 0;
	for(i=0; i<10000; i++)
		if(lcnt[i] != 0){
			s += lcnt[i];
			printf("%d %d\n",i,lcnt[i]);
		}
	printf("total %d\n",s);

}
#endif

// ------- end prof ---------------

#ifdef DEBUG

PRIVATE int traceflag = 0;

extern FILE *FO;
extern int lv;

// !!! debug

int callstk[200], callstkptr = 0;
void pushcallstk(int x){
	callstkptr++;
	callstk[callstkptr] = x;
}
void popcallstk(void){ callstkptr--; }

void dumpcallstk(void){
	int i;
	printf("call stack :");
	for(i=1; i<=callstkptr; i++)
		printf("%d ",callstk[i]);
	nl();
}

PRIVATE void trace(void){
	int i, a, b;
	printf("%5d ",ip);
//	prCode(ip);
	printf(" |");
	a = sp - 10;
	if( a < 0 ) a = 0;

	for(i=a; i<=sp; i++)
		if(i == fp)
			printf("[%d] ",SS[i]);
		else
			printf("%d ",SS[i]);
	printf("\n");
}

void dumpDS(void){
	int i;
	for(i=1; i<10; i++)	printf("%d:%d ",i,M[i]);
	printf("\n");
}

#endif

void dumpstack(void){
	int i, a;

	a = sp - 10;
	if( a < 0 ) a = 0;
	printf("\n> %d: ",a);
	for(i=a; i<=sp; i++)
		if(i == fp)
			printf("[%d] ",SS[i]);
		else
			printf("%d ",SS[i]);
	printf("<%d>\n",TOS);
}

// --------- end prof and debug -------------

void initfile(void){
	int i;
	filep[0] = stdin;
	filep[1] = stdout;
	filep[2] = stderr;
	for(i=3; i<MAXFILE; i++)
		filep[i] = NULL;
}

void error(char *mess){
	printf("runtime error: %s\n",mess);
	exit(0);
}

// get a new file pointer, return 0 if full
PRIVATE int newfp(void){
	int i;
	for(i=3; i<MAXFILE; i++)
		if(filep[i]==NULL) return i;
	return 0;
}

// fgets(fp) and put it in M[ads]
PRIVATE void xgets(FILE *fp, int ads){
	char buf[MAXLBUF], *c;
	int i;
	c = fgets(buf,MAXLBUF,fp);
	if( c != NULL ){
		for(i=0; buf[i] != 0; i++, ads++)
			M[ads] = (int)buf[i] & 255;
		M[ads] = 0;
	}else{
		M[ads] = EOF_CHAR;
		M[ads+1] = 0;
	}
}

// static alloc memory of size a
PRIVATE int xalloc(int a){
	int b;
	b = DS;
	if( DS + a >= MAXDS )
		error("alloc: out of memory");
	DS += a;
	return b;
}
// exec syscall n
PRIVATE void xsys(int n){
	int a, f;
	char name[NAMELEN];

	switch(n){
	case 1: 				// pr int
		printf("%d",TOS);
		break;
	case 2: 				// pr char
		printf("%c",TOS);
		break;
	case 3: 				// get char
		push(getchar());
		break;
	case 4: 				// gets
		xgets(stdin,TOS);
		break;
	case 5:					// fopen fname mode
		a = TOS;
		f = NEXT;			// fname
		pop1;
		strunpack(name,f);
		f = newfp();
		if(f == 0)
			error("cannot open, too many files");
		switch(a){			// mode
		case 0: filep[f] = fopen(name,"r"); break;
		case 1: filep[f] = fopen(name,"w"); break;
		case 2: filep[f] = fopen(name,"rw"); break;
		default: error("unknown file mode");
		}
		if(filep[f] == NULL)
			error("cannot open file");
		TOS = f;			// return file pointer
		break;
	case 6:					// fclose fp
		a = TOS;
		fclose(filep[a]);
		filep[a] = NULL;
		break;
	case 7:					// fprint fp num
		a = TOS;
		f = NEXT;
		pop1;
		fprintf(filep[f],"%d",a);
		break;
	case 8:					// fprintc fp c
		a = TOS;
		f = NEXT;
		pop1;
		fprintf(filep[f],"%c",a);
		break;
	case 9:
		TOS = fgetc(filep[TOS]); // fgetc fp
		break;
	case 10:				// fgets fp inbuf
		a = TOS;
		f = NEXT;
		pop1;
		xgets(filep[f],a);
		break;
/*
	case 11:				// loadfile
		strunpack(name,TOS);// file name in som-string
		loadfile(name);
		break;
*/

	case 12:				// eval im line from som
		user_cs();
		eval(TOS);
		sys_cs();
		break;

//	case 13: runflag = 0;  	// halt, this is in eval()

	case 14:
		TOS = xalloc(TOS);
		break;
	case 15:				// exec load from user space
		sys_cs();
		eval(load_ads);
		user_cs();
		break;

	default: error("undefined syscall");
	}
}

// exec syscall n
PRIVATE void xsys0(int n){
	switch(n){
	case 3: 				// get char
		TOS = getchar();
		break;
	default: error("undefined syscall");
	}
}

void eval(int ref){
	int ip, op, a, runflag;
	int b, low, hi;

	ip = ref;
	runflag = 1;
	while(runflag ) {
//		cnt++;
//		IP = ip;
//		if(cnt-- < 0) break;
		op = xop[ip];
		a = xarg[ip];
		ip++;

		#ifdef PROF
			stat();
		#endif
// ---------- exec ---------

	switch(op){

	case icAdd: TOS = NEXT + TOS; pop1; break;
	case icSub: TOS = NEXT - TOS; pop1; break;
	case icMul: TOS = NEXT * TOS; pop1; break;
	case icDiv: TOS = NEXT / TOS; pop1; break;
	case icBand:TOS = NEXT & TOS; pop1; break;
	case icBor: TOS = NEXT | TOS; pop1; break;
	case icBxor:TOS = NEXT ^ TOS; pop1; break;
	case icMod: TOS = NEXT % TOS; pop1; break;
	case icEq:  TOS = NEXT == TOS; pop1; break;
	case icNe:  TOS = NEXT != TOS; pop1; break;
	case icLt:  TOS = NEXT <  TOS; pop1; break;
	case icLe:  TOS = NEXT <= TOS; pop1; break;
	case icGt:  TOS = NEXT >  TOS; pop1; break;
	case icGe:  TOS = NEXT >= TOS; pop1; break;
	case icShl: TOS = NEXT << TOS; pop1; break;
	case icShr: TOS = NEXT >> TOS; pop1; break;

	case icAddv: TOS += ARG; break;
	case icSubv: TOS -= ARG; break;
	case icMulv: TOS *= ARG; break;
	case icDivv: TOS /= ARG; break;
	case icBandv:TOS &= ARG; break;
	case icBorv: TOS |= ARG; break;
	case icBxorv:TOS ^= ARG; break;
	case icModv: TOS %= ARG; break;
	case icEqv:  TOS = TOS == ARG; break;
	case icNev:  TOS = TOS != ARG; break;
	case icLtv:  TOS = TOS <  ARG; break;
	case icLev:  TOS = TOS <= ARG; break;
	case icGtv:  TOS = TOS >  ARG; break;
	case icGev:  TOS = TOS >= ARG; break;
	case icShlv: TOS <<= ARG; break;
	case icShrv: TOS >>= ARG; break;

	case icAddi: TOS += a; break;
	case icSubi: TOS -= a; break;
	case icMuli: TOS *= a; break;
	case icDivi: TOS /= a; break;
	case icBandi:TOS &= a; break;
	case icBori: TOS |= a; break;
	case icBxori:TOS ^= a; break;
	case icModi: TOS %= a; break;
	case icEqi:  TOS = TOS == a; break;
	case icNei:  TOS = TOS != a; break;
	case icLti:  TOS = TOS <  a; break;
	case icLei:  TOS = TOS <= a; break;
	case icGti:  TOS = TOS >  a; break;
	case icGei:  TOS = TOS >= a; break;
	case icShli: TOS <<= a; break;
	case icShri: TOS >>= a; break;

	case icGet: push(ARG); break;
	case icPut: ARG = pop(); break;
	case icLd:	push(M[a]); break;
	case icSt:	M[a] = TOS; break;
	case icLdx: TOS = M[TOS+ARG]; break;
	case icLdy: TOS = M[TOS + M[a]]; break;
	case icStx: M[NEXT+ARG] = TOS; pop1; break;
	case icSty: M[NEXT + M[a]] = TOS; pop1; break;

	case icJmp:	ip = a;	break;
	case icJf: if(TOS == FALSE) ip = a; break;
	case icJt: if(TOS != FALSE) ip = a;	break;

	case icCall:
		sp++;
		SS[sp] = tos;		// push tos
		if (sp > MAXSS)
			error("stack overflow");
		b = xarg[a];
		SS[sp+b] = fp;
		fp = sp + b;
		sp = fp + 1;
		SS[sp] = ip;
		ip = a + 1;
//		pushcallstk(a);		// !!! debug
		F = &SS[fp];
//		dumpstack();
		break;

	case icCall0:			// do not flush tos
		if (sp > MAXSS)
			error("stack overflow");
		b = xarg[a];
		SS[sp+b] = fp;
		fp = sp + b;
		sp = fp + 1;
		SS[sp] = ip;
		ip = a + 1;
		F = &SS[fp];
		break;

	case icRet:				// return tos, if any
		ip = SS[fp+1];
		sp = fp - a;
		fp = SS[fp];
		F = &SS[fp];
		break;

	case icEfor:
		ARG++;
		if(ARG <= F[a+1])
			ip = xarg[ip];	// flow through jt.x
		else
			ip++;			// skip over jt.x
		break;

	case icInc: ARG++; break;
	case icDec: ARG--; break;
	case icLit:
	case icAds:	push(a); break;
	case icSys:
		if(a == 13) runflag = 0; else xsys(a);
		break;

	case icNot: TOS = TOS == FALSE; break;
	case icCase:
		low = xarg[ip];
		hi = xarg[ip+1];
		a = TOS;
		if( a < low || a > hi )
			ip += 2;
		else
			ip += a-low+3;
		break;

	case icEnd:	runflag = 0; break;

	case icGet0: TOS = ARG; break;
	case icPut0: ARG = TOS; break;
	case icLd0:	TOS = M[a]; break;
	case icLit0:
	case icAds0: TOS = a; break;
	case icSys0: xsys0(a); break;

	default:
		error("undefined opcode");
	}
// ----- end exec ----
//	dumpstack();
	}
}

